PROGRAM Cavidad 

    !*******************   
    !CODIGO BY VUKODLAK*
	!*******************
   
   IMPLICIT NONE
   
   INTEGER::  i , j , nmed , ncuar ,n1 ,n3, itera
   INTEGER, PARAMETER:: N=99
   REAL*8:: h , h2 , h3, hcuad , error_psi , error_chi
   REAL*8, PARAMETER:: Re=1. , tol=0.0001, alfa=1.5
   REAL*8, ALLOCATABLE:: psi(:,:) , psi_old(:,:) , chi(:,:) , &
                         chi_old(:,:), u(:,:), v(:,:)
   CHARACTER, PARAMETER:: lectura = "n"
 
 !HPF$ DISTRIBUTE(BLOCK,BLOCK)::psi, psi_old, chi, chi_old, u, v

 !ESTE PROGRAMA RESUELVE LAS ECUACIONES DE NAVIER-STOKES USANDO LA FORMULACION VORTICIDAD
 !CORRIENTE PARA UNA CAVIDAD DONDE EL FLUJO ENTRA POR UN ORIFICIO Y SALE POR OTRO. 
 !EL DOMINIO DE NUESTRO PROBLEMA SERA:
 !		   
 !	 (0,1) 	________________________   (1,1)
 !		   |						|		     N    = NUMERO DE NODOS INTERIORES DEL DOMINIO
 !		   |		   C		D	|			 Re   = NUMERO DE REYNOLDS
 !		   |						|  (1,3/4)	 psi  = FUNCION DE CORRIENTE
 !		   |						 --			 chi  = VORTICIDAD
 !		   |						 --	u_0		 alfa = PARAMETRO DE RELAJACION
 !		   |  B						 -- 
 !		   |						|  (1,1/2)
 !		   |						|
 !		   |						|
 !		   |					F	|
 !		   |						|
 !		   |  A	 			G		|
 !	 (0,0) |_____      _____________|  (1,0)
 !				  |	| |
 !         		   u_0
 !          (1/4,0)       (1/2,0)	 			   
  
!ASIGNACION DE LAS DIMENSIONES DE LAS MATRICES psi(FUNCION DE CORRIENTE) Y chi(VORTICIDAD) 
   
       ALLOCATE ( psi(0:N+1,0:N+1) , psi_old(0:N+1,0:N+1) , &
                  chi(0:N+1,0:N+1) , chi_old(0:N+1,0:N+1) , &
			      u(0:N+1,0:N+1)   , v(0:N+1,0:N+1) )
	 
!ALGUNOS PARAMETROS IMPORTANTES
   
       n1      = N+1
       n3      = 3*(N+1)/4
       nmed    = (N+1)/2
       ncuar   = (N+1)/4
       h       = 1./(N+1)
       h2      = 2./(h*h)
       h3      = 1./2*h
       hcuad   = h*h
       
!CONDICIONES DE CONTORNO
!PARA psi(FUNCION DE CORRIENTE) CON SECCIONES MATRICIALES

       psi(0:ncuar,0)  = 0              !Pared A
       psi(0,:)        = 0              !Pared B
       psi(:,n1)       = 0              !Pared C
       psi(n1,n3:n1)   = 0              !Pared D
       psi(n1,0:nmed)  = 1/4.           !Pared F
       psi(nmed:n1,0)  = 1/4.           !Pared G

	   DO j=nmed,n3
  	      psi(n1,j) = (3/4.) - j * h    !Pared E
	   END DO
	
	   DO i=ncuar,nmed
	      psi(i,0) = i * h - (1/4.)     !Pared H
	   END DO
   	
!CONDICIONES INICIALES PARA psi(FUNCION DE CORRIENTE) Y chi(VORTICIDAD)

    chi          = 0
    psi(1:N,1:N) = 0
	itera        = 0

    IF (lectura=="s") THEN
       OPEN (UNIT=28,FILE="inicio")
          DO i=0,N+1
            DO j=0,N+1
                READ(28,*) psi(i,j),chi(i,j)
            END DO	
          END DO
    END IF

!*************************************************************************!
!ALGORITMO DE RESOLUCION PARA LAS DOS ECUACIONES ACOPLADAS ENTRE psi Y chi!  
!*************************************************************************! 

	  DO  								 
   		  itera   = itera + 1  !NUMERO DE ITERACIONES
          chi_old = chi        !ACTUALIZACION DE MATRICES chi Y psi
          psi_old = psi
   
!1 ITERACIONES CON JACOBI+RED-BLACK+SOR PARA psi CON BUCLES
	       
!HPF$ INDEPENDENT,	NEW(j)
      DO i=1,n
!HPF$ INDEPENDENT	
         DO j=1,n
            IF (MOD(i+j,2).EQ.0) THEN
               psi(i,j) = psi_old(i,j) + alfa * (0.25      &
                        * (psi_old(i-1,j) + psi_old(i+1,j) &
                        + psi_old(i,j-1) + psi_old(i,j+1)  &
                        - hcuad * chi(i,j)) - psi_old(i,j))
			END IF
         END DO
      END DO
    
  psi_old=psi !ACTUALIZACION DE LOS NODOS PARES
		
!HPF$ INDEPENDENT,	NEW(j)
      DO i=1,n
!HPF$ INDEPENDENT
         DO j=1,n
            IF (MOD(i+j,2).NE.0) THEN 
		       psi(i,j) = psi_old(i,j) + alfa * (0.25 	   &
			            * (psi_old(i-1,j) + psi_old(i+1,j) &
                        + psi_old(i,j-1) + psi_old(i,j+1)  &
                        - hcuad * chi(i,j)) - psi_old(i,j))
            END IF
         END DO
	  END DO

!CONDICIONES DE CONTORNO PARA chi (VORTICIDAD)

          chi(1:N,0)  = h2 * (psi(1:N,1) - psi(1:N,0))   !Pared de Abajo
          chi(1:N,n1) = h2 * (psi(1:N,N) - psi(1:N,n1))  !Pared de Arriba
          chi(0,1:N)  = h2 * (psi(1,1:N) - psi(0,1:N))   !Pared Izquierda
          chi(n1,1:N) = h2 * (psi(N,1:N) - psi(n1,1:N))  !Pared Derecha

!1 ITERACION CON RED-BLACK+SOR PARA chi CON BUCLES

!HPF$ INDEPENDENT,	NEW(j)
      DO i=1,n
!HPF$ INDEPENDENT
         DO j=1,n
    	    IF (MOD(i+j,2).EQ.0) THEN
		       chi(i,j) = chi_old(i,j) + alfa * (0.25         & 
			            * (chi_old(i-1,j) + chi_old(i+1,j)    &
                        + chi_old(i,j-1) + chi_old(i,j+1))    &
                        - chi_old(i,j)) - 0.0625 * Re * alfa  &
                        * ((psi(i,j+1) - psi(i,j-1)) 		  &
						* (chi_old(i+1,j) - chi_old(i-1,j))   &
                        - (psi(i+1,j) - psi(i-1,j)) 		  &
						* (chi_old(i,j+1) - chi_old(i,j-1)))
			END IF
	     END DO
       END DO

	chi_old=chi	  !ACTUALIZAMOS LOS NODOS PARES

!HPF$ INDEPENDENT,	NEW(j)
      DO i=1,n
!HPF$ INDEPENDENT
	     DO j=1,n
            IF (MOD(i+j,2).NE.0) THEN
		       chi(i,j) = chi_old(i,j) + alfa * (0.25 		 &
			            * (chi_old(i-1,j) + chi_old(i+1,j)	 &
						+ chi_old(i,j-1) + chi_old(i,j+1))	 &
						- chi_old(i,j)) - 0.0625 * Re 		 &
						* alfa * ((psi(i,j+1) - psi(i,j-1))	 &
						* (chi_old(i+1,j) - chi_old(i-1,j))	 &
						- (psi(i+1,j) - psi(i-1,j)) 		 &
						* (chi_old(i,j+1) - chi_old(i,j-1))) 
		    END IF
	     END DO
      END DO

!CALCULAMOS NORMA EN L2 DE LOS RESIDUOS

	      error_psi = 0
	      error_chi = 0

!HPF$ INDEPENDENT, REDUCTION(error_psi,error_chi)
 
      DO j=1,n
         DO i=1,n
            error_psi =  error_psi + (psi(i,j) - psi_old(i,j)) &
			          *  (psi(i,j) - psi_old(i,j))
	        error_chi =  error_chi + (chi(i,j) -  chi_old(i,j)) &
		              *  (chi(i,j) - chi_old(i,j))
         END DO
	  END DO

	  error_psi = SQRT(error_psi)
	  error_chi = SQRT(error_chi)

   PRINT*,"iteracion=",itera,"   error_psi= ",error_psi,"   error_chi= ",error_chi

!COMPROBAMOS SI SE ALCANZA LA TOLERANCIA
	
	    IF ((error_psi<=tol).AND.(error_chi<=tol)) EXIT	 
		!SALDREMOS DEL BUCLE CUANDO SE ALCANCEN AMBAS TOLERANCIAS

      END DO
	  
!*************************************************************************!
!FINAL DEL ALGORITMO DE RESOLUCION                                        !  
!*************************************************************************!
	
!CALCULAMOS LAS COMPONENTES DE LA VELOCIDAD u Y v
	  
	  	 v(1:N,1:N) = h3 * (psi(0:N-1,1:N) - psi(2:n1,1:N)) 
	  	 u(1:N,1:N) = h3 * (psi(1:N,2:n1)  - psi(1:N,0:N-1))
	     
!ALMACENAMOS LOS DATOS PARA ANALIZAR RESULTADOS CON MATLAB

		 DO j=0,N+1
           DO i=0,N+1
            WRITE(9,999) i,j,psi(i,j),u(i,j),v(i,j)
           END DO
         END DO

         DO j=0,n+1
           DO i=0,n+1
            WRITE(10,99) i,j,chi(i,j)
           END DO
         END DO

         99 FORMAT (1x,i4,1x,i4,2x,E10.4)
         999 FORMAT (1x,i4,1x,i4,3(2x,E10.4))

!ALMACENAMOS LOS DATOS PARA psi Y chi

     DO i=0,N+1
        DO j=0,N+1
            WRITE(28,*) psi(i,j), chi(i,j)
        END DO
     END DO

END PROGRAM Cavidad








